#include <stdio.h>
#include <memory.h>

#include "Cpu_68k.h"
#include "M68KD.h"
#include "Mem_M68k.h"
#include "Mem_S68k.h"

#define uint32 unsigned int

FILE* trace;
FILE* fp_call;
extern FILE* fp_hook;
extern bool rd_mode, wr_mode, ppu_mode, pc_mode;
extern uint32 rd_low,rd_high;
extern uint32 wr_low,wr_high;
extern uint32 ppu_low,ppu_high;
extern uint32 pc_low,pc_high;
extern uint32 dump_low;
extern bool trace_map;
extern bool hook_trace;

extern "C" {
	uint32 hook_address;
	uint32 hook_value;
	uint32 hook_pc;
	
	void hook_read_byte();
	void hook_read_word();
	void hook_read_dword();
	void hook_write_byte();
	void hook_write_word();
	void hook_write_dword();

	void GensTrace();
};

static bool *mapped=0;
uint32 Current_PC;
int Debug = 1;


void DeInitDebug()
{
  if( mapped ) delete[] mapped;
	if( trace ) fclose( trace );
}

unsigned short Next_Word_T(void)
{
	unsigned short val;
	
	if (Debug == 1) val = M68K_RW(Current_PC);
	else if (Debug >= 2) val = S68K_RW(Current_PC);

	Current_PC += 2;

	return(val);
}

unsigned int Next_Long_T(void)
{
	unsigned int val;
	
	if (Debug == 1)
	{
		val = M68K_RW(Current_PC);
		val <<= 16;
		val |= M68K_RW(Current_PC + 2);
	}
	else if (Debug >= 2)
	{
		val = S68K_RW(Current_PC);
		val <<= 16;
		val |= S68K_RW(Current_PC + 2);
	}

	Current_PC += 4;

	return(val);
}

void GensTrace()
{
  if( !trace ) {
    trace = fopen( "trace.log", "w" );
    mapped = new bool[ 0x100*0x10000 ];
    memset( mapped,0,0x100*0x10000 );
  }

	if( !hook_trace && !trace_map ) return;
  if( trace_map && mapped[ hook_pc ] )
    return;
	else if( hook_trace && ( hook_pc < pc_low || hook_pc > pc_high ) )
    return;

	char String [512];
	Current_PC = hook_pc;

	int PC;
	int OPC = M68K_RW(Current_PC);

	PC = Current_PC;
	sprintf( String, "%02X:%04X  %02X %02X  %-33s",
		PC >> 16, PC & 0xffff, OPC >> 8, OPC & 0xff,
		M68KDisasm( Next_Word_T, Next_Long_T ) );
	fprintf( trace, "%s", String );

	sprintf( String, "A0=%.8X A1=%.8X A2=%.8X ", main68k_context.areg[0], main68k_context.areg[1], main68k_context.areg[2]);
	fprintf( trace, "%s", String );

	sprintf( String, "A3=%.8X A4=%.8X A5=%.8X ", main68k_context.areg[3], main68k_context.areg[4], main68k_context.areg[5]);
	fprintf( trace, "%s", String );

	sprintf( String, "A6=%.8X A7=%.8X D0=%.8X ", main68k_context.areg[6], main68k_context.areg[7], main68k_context.dreg[0]);
	fprintf( trace, "%s", String );

	sprintf( String, "D1=%.8X D2=%.8X D3=%.8X ", main68k_context.dreg[1], main68k_context.dreg[2], main68k_context.dreg[3]);
	fprintf( trace, "%s", String );

	sprintf( String, "D4=%.8X D5=%.8X D6=%.8X ", main68k_context.dreg[4], main68k_context.dreg[5], main68k_context.dreg[6]);
	fprintf( trace, "%s", String );

	sprintf( String, "D7=%.8X ", main68k_context.dreg[7]);
	fprintf( trace, "%s", String );

	fprintf( trace, "%c", (main68k_context.sr & 0x10)?'X':'x' );
	fprintf( trace, "%c", (main68k_context.sr & 0x08)?'N':'n' );
	fprintf( trace, "%c", (main68k_context.sr & 0x04)?'Z':'z' );
	fprintf( trace, "%c", (main68k_context.sr & 0x02)?'V':'v' );
	fprintf( trace, "%c", (main68k_context.sr & 0x01)?'C':'c' );

	fprintf( trace, "\n", String );

  mapped[ hook_pc ] = true;
}


void hook_read_byte()
{
	if( !hook_trace ) return;

	uint32 Address = hook_address;

	if( rd_mode ) {
		// contiguous
		if( rd_low <= Address && Address <= rd_high )
			fprintf( fp_hook, "[%02X:%04X] %s [%06X]\n",
				hook_pc >> 16, hook_pc & 0xffff, "R08", Address );
	}
	else {
		int bank = Address >> 16;
		int offset = Address & 0xffff;

		int rd_low_bank = rd_low >> 16;
		int rd_low_offset = rd_low & 0xffff;
		int rd_high_bank = rd_high >> 16;
		int rd_high_offset = rd_high & 0xffff;

		// shadow
		if( rd_low_bank <= bank && bank <= rd_high_bank )
		if( rd_low_offset <= offset && offset <= rd_high_offset )
			fprintf( fp_hook, "[%02X:%04X] %s [%06X]\n",
				hook_pc >> 16, hook_pc & 0xffff, "R08", Address );
	}
}

void hook_read_word()
{
	if( !hook_trace ) return;

	uint32 Address = hook_address;

	if( rd_mode ) {
		// contiguous
		if( ( rd_low & ~0x01 ) <= Address && Address <= ( rd_high | 0x01 ) )
			fprintf( fp_hook, "[%02X:%04X] %s [%06X]\n",
				hook_pc >> 16, hook_pc & 0xffff, "R16", Address );
	}
	else {
		int bank = Address >> 16;
		int offset = Address & 0xffff;

		int rd_low_bank = rd_low >> 16;
		int rd_low_offset = rd_low & 0xffff;
		int rd_high_bank = rd_high >> 16;
		int rd_high_offset = rd_high & 0xffff;

		// shadow
		if( rd_low_bank <= bank && bank <= rd_high_bank )
		if( ( rd_low_offset & ~0x01 ) <= offset && offset <= ( rd_high_offset | 0x01 ) )
			fprintf( fp_hook, "[%02X:%04X] %s [%06X]\n",
				hook_pc >> 16, hook_pc & 0xffff, "R16", Address );
	}
}

void hook_read_dword()
{
	if( !hook_trace ) return;

	uint32 Address = hook_address;

	if( rd_mode ) {
		// contiguous
		if( ( rd_low & ~0x03 ) <= Address && Address <= ( rd_high | 0x03 ) )
			fprintf( fp_hook, "[%02X:%04X] %s [%06X]\n",
				hook_pc >> 16, hook_pc & 0xffff, "R32", Address );
	}
	else {
		int bank = Address >> 16;
		int offset = Address & 0xffff;

		int rd_low_bank = rd_low >> 16;
		int rd_low_offset = rd_low & 0xffff;
		int rd_high_bank = rd_high >> 16;
		int rd_high_offset = rd_high & 0xffff;

		// shadow
		if( rd_low_bank <= bank && bank <= rd_high_bank )
		if( ( rd_low_offset & ~0x03 ) <= offset && offset <= ( rd_high_offset | 0x03 ) )
			fprintf( fp_hook, "[%02X:%04X] %s [%06X]\n",
				hook_pc >> 16, hook_pc & 0xffff, "R32", Address );
	}
}

void hook_write_byte()
{
	if( !hook_trace ) return;

	uint32 Address = hook_address;
	uint32 Value = hook_value;

	if( wr_mode ) {
		// contiguous
		if( wr_low <= Address && Address <= wr_high )
			fprintf( fp_hook, "[%02X:%04X] %s = %02X   [%06X]\n",
				hook_pc >> 16, hook_pc & 0xffff, "W08", Value, Address );
	}
	else {
		int bank = Address >> 16;
		int offset = Address & 0xffff;
		int wr_low_bank = wr_low >> 16;
		int wr_low_offset = wr_low & 0xffff;
		int wr_high_bank = wr_high >> 16;
		int wr_high_offset = wr_high & 0xffff;

		// shadow
		if( wr_low_bank <= bank && bank <= wr_high_bank )
		if( wr_low_offset <= offset && offset <= wr_high_offset )
			fprintf(fp_hook, "[%02X:%04X] %s = %02X   [%06X]\n",
				hook_pc >> 16, hook_pc & 0xffff, "W08", Value, Address);
	}
}

void hook_write_word()
{
	if( !hook_trace ) return;

	uint32 Address = hook_address;
	uint32 Value = hook_value;

	if( wr_mode ) {
		// contiguous
		if( ( wr_low & ~0x01 ) <= Address && Address <= ( wr_high | 0x01 ) )
			fprintf(fp_hook, "[%02X:%04X] %s = %04X [%06X]\n",
				hook_pc >> 16, hook_pc & 0xffff, "W16", Value, Address );
	}
	else {
		int bank = Address >> 16;
		int offset = Address & 0xffff;
		int wr_low_bank = wr_low >> 16;
		int wr_low_offset = wr_low & 0xffff;
		int wr_high_bank = wr_high >> 16;
		int wr_high_offset = wr_high & 0xffff;

		// shadow
		if( wr_low_bank <= bank && bank <= wr_high_bank )
		if( ( wr_low_offset & ~0x01 ) <= offset && offset <= ( wr_high_offset | 0x01 ) )
			fprintf( fp_hook, "[%02X:%04X] %s = %04X [%06X]\n",
				hook_pc >> 16, hook_pc & 0xffff, "W16", Value, Address );
	}
}

void hook_write_dword()
{
	if( !hook_trace ) return;

	uint32 Address = hook_address;
	uint32 Value = hook_value;

	if( wr_mode ) {
		// contiguous
		if( ( wr_low & ~0x03 ) <= Address && Address <= ( wr_high | 0x03 ) )
			fprintf(fp_hook, "[%02X:%04X] %s = %08X [%06X]\n",
				hook_pc >> 16, hook_pc & 0xffff, "W32", Value, Address );
	}
	else {
		int bank = Address >> 16;
		int offset = Address & 0xffff;
		int wr_low_bank = wr_low >> 16;
		int wr_low_offset = wr_low & 0xffff;
		int wr_high_bank = wr_high >> 16;
		int wr_high_offset = wr_high & 0xffff;

		// shadow
		if( wr_low_bank <= bank && bank <= wr_high_bank )
		if( ( wr_low_offset & ~0x03 ) <= offset && offset <= ( wr_high_offset | 0x03 ) )
			fprintf( fp_hook, "[%02X:%04X] %s = %08X [%06X]\n",
				hook_pc >> 16, hook_pc & 0xffff, "W32", Value, Address );
	}
}